"""scrapli.ffi_types"""

from ctypes import (
    CFUNCTYPE,
    POINTER,
    Structure,
    c_bool,
    c_char_p,
    c_int,
    c_size_t,
    c_uint,
    c_uint8,
    c_uint64,
    c_void_p,
    cast,
)
from typing import TypeAlias

DriverPointer = c_void_p
OperationId = c_uint

# mypy seems to dislike the pointer bits, but these do accurately reflect the api surface,
# so... we'll just tell mypy to chill out on these
OperationIdPointer: TypeAlias = POINTER(OperationId)  # type: ignore[valid-type]
CancelPointer: TypeAlias = POINTER(c_bool)  # type: ignore[valid-type]
ZigSlicePointer: TypeAlias = POINTER("ZigSlice")  # type: ignore[valid-type, call-overload]
StringPointer: TypeAlias = POINTER(c_char_p)  # type: ignore[valid-type]
IntPointer: TypeAlias = POINTER(c_int)  # type: ignore[valid-type]
U64Pointer: TypeAlias = POINTER(c_uint64)  # type: ignore[valid-type]
BoolPointer: TypeAlias = POINTER(c_bool)  # type: ignore[valid-type]


class ZigU64Slice(Structure):
    """
    A struct representing a slice of u64 in zig.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    # fields must be declared this way to be c types compatible
    _fields_ = [  # noqa: RUF012
        ("ptr", POINTER(c_uint64)),
        ("len", c_size_t),
    ]

    def __init__(self, size: c_int):
        self.ptr = cast((c_uint64 * size.value)(), POINTER(c_uint64))
        self.len = size.value

        super().__init__()

    def get_contents(self) -> list[int]:
        """
        Return the contents of the slice as a list of ints.

        Args:
            N/A

        Returns:
            list[int]: the slice contents

        Raises:
            N/A

        """
        return [self.ptr[i] for i in range(self.len)]


class ZigSlice(Structure):
    """
    A struct representing a slice of u8 (a string) in zig.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    _fields_ = [  # noqa: RUF012
        ("ptr", POINTER(c_uint8)),
        ("len", c_size_t),
        ("value", c_char_p),
    ]

    def __init__(self, size: c_int):
        self.ptr = cast((c_uint8 * size.value)(), POINTER(c_uint8))
        self.len = size.value

        super().__init__()

    def get_contents(self) -> bytes:
        """
        Return the contents of the slice as bytes.

        Args:
            N/A

        Returns:
            bytes: the slice contents

        Raises:
            N/A

        """
        return bytes(cast(self.ptr, POINTER(c_uint8 * self.len)).contents)

    def get_decoded_contents(self) -> str:
        """
        Return the contents of the slice as str.

        Args:
            N/A

        Returns:
            bytes: the slice contents

        Raises:
            N/A

        """
        return self.get_contents().decode()


def to_c_string(s: str) -> c_char_p:
    """
    Accepts a string and converts it to a c_char_p.

    Args:
        s: the string-like thing to convert to a c_char_p

    Returns:
        c_char_p: the converted string

    Raises:
        N/A

    """
    return c_char_p(s.encode(encoding="utf-8"))


LogFuncCallback: TypeAlias = CFUNCTYPE(None, c_uint8, POINTER(ZigSlice))  # type: ignore[valid-type]



"""scrapli.helper"""

from asyncio import get_event_loop
from os import read
from pathlib import Path
from select import select

from scrapli.exceptions import OperationException


def resolve_file(file: str) -> str:
    """
    Resolve file from provided string

    Args:
        file: string path to file

    Returns:
        str: resolved/expanded (if applicable) string path to file

    Raises:
        OperationException: if file cannot be resolved

    """
    if Path(file).is_file():
        return str(Path(file))
    if Path(file).expanduser().is_file():
        return str(Path(file).expanduser())

    raise OperationException(f"path `{file}` could not be resolved")


def wait_for_available_operation_result(fd: int) -> None:
    """
    Wait for the next operation to be complete.

    Args:
        fd: the fd to wait on

    Returns:
        None

    Raises:
        N/A

    """
    _, _, _ = select([fd], [], [])
    read(fd, 1)


async def _wait_for_fd_readable(fd: int) -> None:
    """
    Wait for fd to be readable.

    Args:
        fd: the fd to wait on

    Returns:
        None

    Raises:
        N/A

    """
    loop = get_event_loop()

    fut = loop.create_future()

    def on_ready() -> None:
        loop.remove_reader(fd)

        fut.set_result(None)

    loop.add_reader(fd, on_ready)

    await fut


async def wait_for_available_operation_result_async(fd: int) -> None:
    """
    Wait for the next operation to be complete.

    Args:
        fd: the fd to wait on

    Returns:
        None

    Raises:
        N/A

    """
    await _wait_for_fd_readable(fd)

    read(fd, 1)



"""scrapli.session"""

from ctypes import c_char_p, c_uint64
from dataclasses import dataclass, field

from scrapli.exceptions import OptionsException
from scrapli.ffi_mapping import LibScrapliMapping
from scrapli.ffi_types import DriverPointer, to_c_string


@dataclass
class Options:
    """
    Options holds session related options to pass to the ffi layer.

    Args:
        read_size: read size
        return_char: return char
        operation_timeout_s: operation timeout in s, ignored if operation_timeout_ns set
        operation_timeout_ns: operation timeout in ns
        operation_max_search_depth: operation maximum search depth
        recorder_path: path for session recorder output to write to

    Returns:
        None

    Raises:
        N/A

    """

    read_size: int | None = None
    return_char: str | None = None
    operation_timeout_s: int | None = None
    operation_timeout_ns: int | None = None
    operation_max_search_depth: int | None = None
    recorder_path: str | None = None

    _return_char: c_char_p | None = field(init=False, default=None, repr=False)
    _recorder_path: c_char_p | None = field(init=False, default=None, repr=False)

    def __post_init__(self) -> None:
        if self.operation_timeout_s is not None or self.operation_timeout_ns is not None:
            if self.operation_timeout_ns is None and self.operation_timeout_s is not None:
                self.operation_timeout_ns = int(self.operation_timeout_s / 1e-9)

    def apply(self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer) -> None:  # noqa: C901
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.read_size is not None:
            status = ffi_mapping.options_mapping.session.set_read_size(
                ptr, c_uint64(self.read_size)
            )
            if status != 0:
                raise OptionsException("failed to set session read size")

        if self.return_char is not None:
            self._return_char = to_c_string(self.return_char)

            status = ffi_mapping.options_mapping.session.set_return_char(ptr, self._return_char)
            if status != 0:
                raise OptionsException("failed to set session return char")

        if self.operation_timeout_ns is not None:
            status = ffi_mapping.options_mapping.session.set_operation_timeout_ns(
                ptr, c_uint64(self.operation_timeout_ns)
            )
            if status != 0:
                raise OptionsException("failed to set session operation timeout")

        if self.operation_max_search_depth is not None:
            status = ffi_mapping.options_mapping.session.set_operation_max_search_depth(
                ptr, c_uint64(self.operation_max_search_depth)
            )
            if status != 0:
                raise OptionsException("failed to set session operation max search depth")

        if self.recorder_path is not None:
            self._recorder_path = to_c_string(self.recorder_path)

            status = ffi_mapping.options_mapping.session.set_recorder_path(ptr, self._recorder_path)
            if status != 0:
                raise OptionsException("failed to set session recorder path")

    def __repr__(self) -> str:
        """
        Magic repr method for Options object

        Args:
            N/A

        Returns:
            str: repr for Options object

        Raises:
            N/A

        """
        return (
            # it will probably be "canonical" to import Options as SessionOptions, so we'll make
            # the repr do that too
            f"Session{self.__class__.__name__}("
            f"read_size={self.read_size!r}, "
            f"return_char={self.return_char!r}) "
            f"operation_timeout_ns={self.operation_timeout_ns!r}) "
            f"operation_max_search_depth={self.operation_max_search_depth!r}) "
            f"recorder_path={self.recorder_path!r}) "
        )



"""scrapli.result"""

from dataclasses import dataclass


@dataclass
class Result:
    """
    Result holds the result of an operation.

    Args:
        input_:
        host:
        port:
        start_time:
        end_time:
        result_raw:
        result:
        rpc_warnings:
        rpc_errors:

    Returns:
        None

    Raises:
        N/A

    """

    input_: str
    host: str
    port: int
    start_time: int
    end_time: int
    result_raw: bytes
    result: str
    rpc_warnings: str
    rpc_errors: str

    @property
    def failed(self) -> bool:
        """
        Returns True if any failed indicators were seen, otherwise False.

        Args:
            N/A

        Returns:
            bool: True for failed, otherwise False

        Raises:
            N/A

        """
        return bool(self.rpc_errors)

    @property
    def elapsed_time_seconds(self) -> float:
        """
        Returns the number of seconds the operation took.

        Args:
            N/A

        Returns:
            float: duration in seconds

        Raises:
            N/A

        """
        return (self.end_time - self.start_time) / 1_000_000_000



"""scrapli.cli"""

import importlib.resources
from collections.abc import Callable
from ctypes import (
    c_bool,
    c_char_p,
    c_int,
    c_uint,
    c_uint64,
)
from enum import Enum
from logging import getLogger
from os import environ
from pathlib import Path
from types import TracebackType
from typing import Any

from scrapli.auth import Options as AuthOptions
from scrapli.cli_decorators import handle_operation_timeout, handle_operation_timeout_async
from scrapli.cli_result import Result
from scrapli.exceptions import (
    AllocationException,
    CloseException,
    GetResultException,
    NotOpenedException,
    OpenException,
    OperationException,
    OptionsException,
    SubmitOperationException,
)
from scrapli.ffi_mapping import LibScrapliMapping
from scrapli.ffi_types import (
    CancelPointer,
    DriverPointer,
    IntPointer,
    LogFuncCallback,
    OperationIdPointer,
    U64Pointer,
    ZigSlice,
    ZigU64Slice,
    to_c_string,
)
from scrapli.helper import (
    resolve_file,
    wait_for_available_operation_result,
    wait_for_available_operation_result_async,
)
from scrapli.session import Options as SessionOptions
from scrapli.transport import Options as TransportOptions

CLI_DEFINITIONS_PATH_OVERRIDE_ENV = "SCRAPLI_DEFINITIONS_PATH"
CLI_DEFINITIONS_PATH_OVERRIDE = environ.get(CLI_DEFINITIONS_PATH_OVERRIDE_ENV)


class InputHandling(str, Enum):
    """
    Enum representing the input handling flavor.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    EXACT = "exact"
    FUZZY = "fuzzy"
    IGNORE = "ignore"


class Cli:
    """
    Cli represents a cli connection object.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(  # noqa: PLR0913
        self,
        definition_file_or_name: str,
        host: str,
        *,
        logger_callback: Callable[[int, str], None] | None = None,
        port: int = 22,
        auth_options: AuthOptions | None = None,
        session_options: SessionOptions | None = None,
        transport_options: TransportOptions | None = None,
        logging_uid: str | None = None,
    ) -> None:
        logger_name = f"{__name__}.{host}:{port}"
        if logging_uid is not None:
            logger_name += f":{logging_uid}"

        self.logger = getLogger(logger_name)
        self._logging_uid = logging_uid

        self.ffi_mapping = LibScrapliMapping()

        self.definition_file_or_name = definition_file_or_name
        self._load_definition(definition_file_or_name=definition_file_or_name)

        # note: many places have encodings done prior to function calls such that the encoded
        # result is not gc'd prior to being used in zig-land, so it looks a bit weird, but thats
        # why. in this case we also just store the host since thats cheap and we need it as a string
        # in places too
        self.host = host
        self._host = to_c_string(host)

        self.logger_callback = (
            LogFuncCallback(logger_callback) if logger_callback else LogFuncCallback(0)
        )
        self._logger_callback = logger_callback

        self.port = port

        self.auth_options = auth_options or AuthOptions()
        self.session_options = session_options or SessionOptions()
        self.transport_options = transport_options or TransportOptions()

        self.ptr: DriverPointer | None = None
        self.poll_fd: int = 0

        self._ntc_templates_platform: str | None = None
        self._genie_platform: str | None = None

    def __enter__(self: "Cli") -> "Cli":
        """
        Enter method for context manager.

        Args:
            N/A

        Returns:
            Cli: a concrete implementation of the opened Cli object

        Raises:
            N/A

        """
        self.open()

        return self

    def __exit__(
        self,
        exception_type: BaseException | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """
        Exit method to cleanup for context manager.

        Args:
            exception_type: exception type being raised
            exception_value: message from exception being raised
            traceback: traceback from exception being raised

        Returns:
            None

        Raises:
            N/A

        """
        self.close()

    async def __aenter__(self: "Cli") -> "Cli":
        """
        Enter method for context manager.

        Args:
            N/A

        Returns:
            Cli: a concrete implementation of the opened Cli object

        Raises:
            N/A

        """
        await self.open_async()

        return self

    async def __aexit__(
        self,
        exception_type: BaseException | None,
        exception_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """
        Exit method to cleanup for context manager.

        Args:
            exception_type: exception type being raised
            exception_value: message from exception being raised
            traceback: traceback from exception being raised

        Returns:
            None

        Raises:
            N/A

        """
        await self.close_async()

    def __str__(self) -> str:
        """
        Magic str method for Cli object.

        Args:
            N/A

        Returns:
            str: str representation of Cli

        Raises:
            N/A

        """
        return f"scrapli.Cli {self.host}:{self.port}"

    def __repr__(self) -> str:
        """
        Magic repr method for Cli object.

        Args:
            N/A

        Returns:
            str: repr for Cli object

        Raises:
            N/A

        """
        return (
            f"{self.__class__.__name__}("
            f"definition_file_or_name={self.definition_file_or_name!r}, "
            f"host={self.host!r}, "
            f"port={self.port!r}, "
            f"auth_options={self.auth_options!r} "
            f"session_options={self.auth_options!r} "
            f"transport_options={self.transport_options!r}) "
        )

    def __copy__(self, memodict: dict[Any, Any] = {}) -> "Cli":
        # reasonably safely copy of the object... *reasonably*... basically assumes that options
        # will never be mutated during an objects lifetime, which *should* be the case. probably.
        return Cli(
            host=self.host,
            definition_file_or_name=self.definition_file_or_name,
            logger_callback=self._logger_callback,
            port=self.port,
            auth_options=self.auth_options,
            session_options=self.session_options,
            transport_options=self.transport_options,
            logging_uid=self._logging_uid,
        )

    def _load_definition(self, definition_file_or_name: str) -> None:
        if CLI_DEFINITIONS_PATH_OVERRIDE is not None:
            definition_path = f"{CLI_DEFINITIONS_PATH_OVERRIDE}/{definition_file_or_name}.yaml"
        else:
            definitions_path = importlib.resources.files("scrapli.definitions")
            definition_path = f"{definitions_path}/{definition_file_or_name}.yaml"

        if Path(definition_path).exists():
            with open(definition_path, "rb") as f:
                self.definition_string = f.read()

            return

        if Path(definition_file_or_name).exists():
            with open(definition_file_or_name, "rb") as f:
                self.definition_string = f.read()

        raise OptionsException(
            f"definition platform name or filename '{definition_file_or_name}' not found"
        )

    def _ptr_or_exception(self) -> DriverPointer:
        if self.ptr is None:
            raise NotOpenedException

        return self.ptr

    def _alloc(
        self,
    ) -> None:
        ptr = self.ffi_mapping.cli_mapping.alloc(
            definition_string=c_char_p(self.definition_string),
            logger_callback=self.logger_callback,
            host=self._host,
            port=c_int(self.port),
            transport_kind=c_char_p(self.transport_options.get_transport_kind()),
        )
        if ptr == 0:  # type: ignore[comparison-overlap]
            raise AllocationException("failed to allocate cli")

        self.ptr = ptr

        poll_fd = int(
            self.ffi_mapping.shared_mapping.get_poll_fd(
                ptr=self._ptr_or_exception(),
            )
        )
        if poll_fd == 0:
            raise AllocationException("failed to allocate cli")

        self.poll_fd = poll_fd

    def _free(
        self,
    ) -> None:
        self.ffi_mapping.shared_mapping.free(ptr=self._ptr_or_exception())

    @property
    def ntc_templates_platform(self) -> str:
        """
        Returns the ntc templates platform for the cli object.

        Args:
            N/A

        Returns:
            str: the ntc templates platform name

        Raises:
            N/A

        """
        if self._ntc_templates_platform is not None:
            return self._ntc_templates_platform

        ntc_templates_platform = ZigSlice(size=c_int(256))

        status = self.ffi_mapping.cli_mapping.get_ntc_templates_platform(
            ptr=self._ptr_or_exception(),
            ntc_templates_platform=ntc_templates_platform,
        )
        if status != 0:
            raise GetResultException("failed to retrieve ntc templates platform")

        self._ntc_templates_platform = ntc_templates_platform.get_decoded_contents().strip("\x00")

        return self._ntc_templates_platform

    @property
    def genie_platform(self) -> str:
        """
        Returns the genie platform for the cli object.

        Args:
            N/A

        Returns:
            str: the genie platform name

        Raises:
            N/A

        """
        if self._genie_platform is not None:
            return self._genie_platform

        genie_platform = ZigSlice(size=c_int(256))

        status = self.ffi_mapping.cli_mapping.get_ntc_templates_platform(
            ptr=self._ptr_or_exception(),
            ntc_templates_platform=genie_platform,
        )
        if status != 0:
            raise GetResultException("failed to retrieve genie templates platform")

        self._genie_platform = genie_platform.get_decoded_contents().strip("\x00")

        return self._genie_platform

    def _open(
        self,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> c_uint:
        self._alloc()

        self.auth_options.apply(self.ffi_mapping, self._ptr_or_exception())
        self.session_options.apply(self.ffi_mapping, self._ptr_or_exception())
        self.transport_options.apply(self.ffi_mapping, self._ptr_or_exception())

        status = self.ffi_mapping.cli_mapping.open(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            self._free()

            raise OpenException("failed to submit open operation")

        return c_uint(operation_id.contents.value)

    def open(
        self,
    ) -> Result:
        """
        Open the cli connection.

        Args:
            N/A

        Returns:
            None

        Raises:
            OpenException: if the operation fails

        """
        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._open(operation_id=operation_id, cancel=cancel)

        return self._get_result(operation_id=operation_id)

    async def open_async(self) -> Result:
        """
        Open the cli connection.

        Args:
            N/A

        Returns:
            None

        Raises:
            OpenException: if the operation fails

        """
        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._open(operation_id=operation_id, cancel=cancel)

        return await self._get_result_async(operation_id=operation_id)

    def _close(self) -> c_uint:
        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        status = self.ffi_mapping.cli_mapping.close(
            ptr=self._ptr_or_exception(), operation_id=operation_id, cancel=cancel
        )
        if status != 0:
            raise CloseException("submitting close operation")

        return c_uint(operation_id.contents.value)

    def close(
        self,
    ) -> Result:
        """
        Close the cli connection.

        Args:
            N/A

        Returns:
            None

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            CloseException: if the operation fails

        """
        operation_id = self._close()

        result = self._get_result(operation_id=operation_id)

        self._free()

        return result

    async def close_async(
        self,
    ) -> Result:
        """
        Close the cli connection.

        Args:
            N/A

        Returns:
            None

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            CloseException: if the operation fails

        """
        operation_id = self._close()

        result = await self._get_result_async(operation_id=operation_id)

        self._free()

        return result

    def _get_result(
        self,
        operation_id: c_uint,
    ) -> Result:
        wait_for_available_operation_result(self.poll_fd)

        operation_count = IntPointer(c_int())
        inputs_size = IntPointer(c_int())
        results_raw_size = IntPointer(c_int())
        results_size = IntPointer(c_int())
        results_failed_indicator_size = IntPointer(c_int())
        err_size = IntPointer(c_int())

        status = self.ffi_mapping.cli_mapping.fetch_sizes(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            operation_count=operation_count,
            inputs_size=inputs_size,
            results_raw_size=results_raw_size,
            results_size=results_size,
            results_failed_indicator_size=results_failed_indicator_size,
            err_size=err_size,
        )
        if status != 0:
            raise GetResultException("wait operation failed")

        start_time = U64Pointer(c_uint64())
        splits = ZigU64Slice(size=operation_count.contents)

        inputs_slice = ZigSlice(size=inputs_size.contents)
        results_raw_slice = ZigSlice(size=results_raw_size.contents)
        results_slice = ZigSlice(size=results_size.contents)

        results_failed_indicator_slice = ZigSlice(size=results_failed_indicator_size.contents)
        err_slice = ZigSlice(size=err_size.contents)

        status = self.ffi_mapping.cli_mapping.fetch(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            start_time=start_time,
            splits=splits,
            inputs_slice=inputs_slice,
            results_raw_slice=results_raw_slice,
            results_slice=results_slice,
            results_failed_indicator_slice=results_failed_indicator_slice,
            err_slice=err_slice,
        )
        if status != 0:
            raise GetResultException("fetch operation failed")

        err_contents = err_slice.get_decoded_contents()
        if err_contents:
            raise OperationException(err_contents)

        return Result(
            inputs=inputs_slice.get_decoded_contents(),
            host=self.host,
            port=self.port,
            start_time=start_time.contents.value,
            splits=splits.get_contents(),
            results_raw=results_raw_slice.get_contents(),
            results=results_slice.get_decoded_contents(),
            results_failed_indicator=results_failed_indicator_slice.get_decoded_contents(),
            textfsm_platform=self.ntc_templates_platform,
            genie_platform=self.genie_platform,
        )

    async def _get_result_async(
        self,
        operation_id: c_uint,
    ) -> Result:
        await wait_for_available_operation_result_async(fd=self.poll_fd)

        operation_count = IntPointer(c_int())
        inputs_size = IntPointer(c_int())
        results_raw_size = IntPointer(c_int())
        results_size = IntPointer(c_int())
        results_failed_indicator_size = IntPointer(c_int())
        err_size = IntPointer(c_int())

        status = self.ffi_mapping.cli_mapping.fetch_sizes(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            operation_count=operation_count,
            inputs_size=inputs_size,
            results_raw_size=results_raw_size,
            results_size=results_size,
            results_failed_indicator_size=results_failed_indicator_size,
            err_size=err_size,
        )
        if status != 0:
            raise GetResultException("fetch operation sizes failed")

        start_time = U64Pointer(c_uint64())
        splits = ZigU64Slice(size=operation_count.contents)

        inputs_slice = ZigSlice(size=inputs_size.contents)
        results_raw_slice = ZigSlice(size=results_raw_size.contents)
        results_slice = ZigSlice(size=results_size.contents)

        results_failed_indicator_slice = ZigSlice(size=results_failed_indicator_size.contents)
        err_slice = ZigSlice(size=err_size.contents)

        status = self.ffi_mapping.cli_mapping.fetch(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            start_time=start_time,
            splits=splits,
            inputs_slice=inputs_slice,
            results_raw_slice=results_raw_slice,
            results_slice=results_slice,
            results_failed_indicator_slice=results_failed_indicator_slice,
            err_slice=err_slice,
        )
        if status != 0:
            raise GetResultException("fetch operation failed")

        err_contents = err_slice.get_decoded_contents()
        if err_contents:
            raise OperationException(err_contents)

        return Result(
            inputs=inputs_slice.get_decoded_contents(),
            host=self.host,
            port=self.port,
            start_time=start_time.contents.value,
            splits=splits.get_contents(),
            results_raw=results_raw_slice.get_contents(),
            results=results_slice.get_decoded_contents(),
            results_failed_indicator=results_failed_indicator_slice.get_decoded_contents(),
            textfsm_platform=self.ntc_templates_platform,
            genie_platform=self.genie_platform,
        )

    def _enter_mode(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        requested_mode: c_char_p,
    ) -> c_uint:
        status = self.ffi_mapping.cli_mapping.enter_mode(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            requested_mode=requested_mode,
        )
        if status != 0:
            raise SubmitOperationException("submitting get prompt operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def enter_mode(
        self,
        requested_mode: str,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Enter the given mode on the cli connection.

        Args:
            requested_mode: name of the mode to enter
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _requested_mode = to_c_string(requested_mode)

        operation_id = self._enter_mode(
            operation_id=operation_id, cancel=cancel, requested_mode=_requested_mode
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def enter_mode_async(
        self,
        requested_mode: str,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Enter the given mode on the cli connection.

        Args:
            requested_mode: name of the mode to enter
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _requested_mode = to_c_string(requested_mode)

        operation_id = self._enter_mode(
            operation_id=operation_id, cancel=cancel, requested_mode=_requested_mode
        )

        return await self._get_result_async(operation_id=operation_id)

    def _get_prompt(
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
    ) -> c_uint:
        status = self.ffi_mapping.cli_mapping.get_prompt(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
        )
        if status != 0:
            raise SubmitOperationException("submitting get prompt operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def get_prompt(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Get the current prompt of the cli connection.

        Args:
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._get_prompt(operation_id=operation_id, cancel=cancel)

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def get_prompt_async(
        self,
        *,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Get the current prompt of the cli connection.

        Args:
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        operation_id = self._get_prompt(operation_id=operation_id, cancel=cancel)

        return await self._get_result_async(operation_id=operation_id)

    def _send_input(  # noqa: PLR0913
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        input_: c_char_p,
        requested_mode: c_char_p,
        input_handling: c_char_p,
        retain_input: c_bool,
        retain_trailing_prompt: c_bool,
    ) -> c_uint:
        status = self.ffi_mapping.cli_mapping.send_input(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            input_=input_,
            requested_mode=requested_mode,
            input_handling=input_handling,
            retain_input=retain_input,
            retain_trailing_prompt=retain_trailing_prompt,
        )
        if status != 0:
            raise SubmitOperationException("submitting send input operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def send_input(  # noqa: PLR0913
        self,
        input_: str,
        *,
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        retain_input: bool = False,
        retain_trailing_prompt: bool = False,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send an input on the cli connection.

        Args:
            input_: the input to send
            requested_mode: name of the mode to send the input at
            input_handling: how to handle the input
            retain_input: retain the input in the final "result"
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _input = to_c_string(input_)
        _requested_mode = to_c_string(requested_mode)
        _input_handling = to_c_string(input_handling)

        operation_id = self._send_input(
            operation_id=operation_id,
            cancel=cancel,
            input_=_input,
            requested_mode=_requested_mode,
            input_handling=_input_handling,
            retain_input=c_bool(retain_input),
            retain_trailing_prompt=c_bool(retain_trailing_prompt),
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def send_input_async(  # noqa: PLR0913
        self,
        input_: str,
        *,
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        retain_input: bool = False,
        retain_trailing_prompt: bool = False,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send an input on the cli connection.

        Args:
            input_: the input to send
            requested_mode: name of the mode to send the input at
            input_handling: how to handle the input
            retain_input: retain the input in the final "result"
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _input = to_c_string(input_)
        _requested_mode = to_c_string(requested_mode)
        _input_handling = to_c_string(input_handling)

        operation_id = self._send_input(
            operation_id=operation_id,
            cancel=cancel,
            input_=_input,
            requested_mode=_requested_mode,
            input_handling=_input_handling,
            retain_input=c_bool(retain_input),
            retain_trailing_prompt=c_bool(retain_trailing_prompt),
        )

        return await self._get_result_async(operation_id=operation_id)

    @handle_operation_timeout
    def send_inputs(  # noqa: PLR0913
        self,
        inputs: list[str],
        *,
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        retain_input: bool = False,
        retain_trailing_prompt: bool = False,
        stop_on_indicated_failure: bool = True,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send inputs (plural!) on the cli connection.

        Args:
            inputs: the inputs to send
            requested_mode: name of the mode to send the input at`
            input_handling: how to handle the input
            retain_input: retain the input in the final "result"
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            stop_on_indicated_failure: stops sending inputs at first indicated failure
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator; note that the timeout here is for the whole operation,
        # meaning all the "inputs" combined, not individually
        _ = operation_timeout_ns

        cancel = CancelPointer(c_bool(False))

        result: Result | None = None

        for input_ in inputs:
            operation_id = OperationIdPointer(c_uint(0))

            _input = to_c_string(input_)
            _requested_mode = to_c_string(requested_mode)
            _input_handling = to_c_string(input_handling)

            operation_id = self._send_input(
                operation_id=operation_id,
                cancel=cancel,
                input_=_input,
                requested_mode=_requested_mode,
                input_handling=_input_handling,
                retain_input=c_bool(retain_input),
                retain_trailing_prompt=c_bool(retain_trailing_prompt),
            )

            _result = self._get_result(operation_id=operation_id)

            if result is None:
                result = _result
            else:
                result.extend(result=_result)

            if result.failed and stop_on_indicated_failure:
                return result

        return result  # type: ignore[return-value]

    @handle_operation_timeout_async
    async def send_inputs_async(  # noqa: PLR0913
        self,
        inputs: list[str],
        *,
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        retain_input: bool = False,
        retain_trailing_prompt: bool = False,
        stop_on_indicated_failure: bool = True,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send inputs (plural!) on the cli connection.

        Args:
            inputs: the inputs to send
            requested_mode: name of the mode to send the input at`
            input_handling: how to handle the input
            retain_input: retain the input in the final "result"
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            stop_on_indicated_failure: stops sending inputs at first indicated failure
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            MultiResult: a MultiResult object representing the operations

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator; note that the timeout here is for the whole operation,
        # meaning all the "inputs" combined, not individually
        _ = operation_timeout_ns

        cancel = CancelPointer(c_bool(False))

        result: Result | None = None

        for input_ in inputs:
            operation_id = OperationIdPointer(c_uint(0))

            _input = to_c_string(input_)
            _requested_mode = to_c_string(requested_mode)
            _input_handling = to_c_string(input_handling)

            operation_id = self._send_input(
                operation_id=operation_id,
                cancel=cancel,
                input_=_input,
                requested_mode=_requested_mode,
                input_handling=_input_handling,
                retain_input=c_bool(retain_input),
                retain_trailing_prompt=c_bool(retain_trailing_prompt),
            )

            _result = await self._get_result_async(operation_id=operation_id)
            if result is None:
                result = _result
            else:
                result.extend(result=_result)

            if result.failed and stop_on_indicated_failure:
                return result

        return result  # type: ignore[return-value]

    def send_inputs_from_file(  # noqa: PLR0913
        self,
        f: str,
        *,
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        retain_input: bool = False,
        retain_trailing_prompt: bool = False,
        stop_on_indicated_failure: bool = True,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send inputs (plural! from a file, line-by-line) on the cli connection.

        Args:
            f: the file to load -- each line will be sent, tralining newlines ignored
            requested_mode: name of the mode to send the input at`
            input_handling: how to handle the input
            retain_input: retain the input in the final "result"
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            stop_on_indicated_failure: stops sending inputs at first indicated failure
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            MultiResult: a MultiResult object representing the operations

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        with open(resolve_file(f), encoding="utf-8", mode="r") as _f:
            inputs = _f.read().splitlines()

        return self.send_inputs(
            inputs=inputs,
            requested_mode=requested_mode,
            input_handling=input_handling,
            retain_input=retain_input,
            retain_trailing_prompt=retain_trailing_prompt,
            stop_on_indicated_failure=stop_on_indicated_failure,
            operation_timeout_ns=operation_timeout_ns,
        )

    async def send_inputs_from_file_async(  # noqa: PLR0913
        self,
        f: str,
        *,
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        retain_input: bool = False,
        retain_trailing_prompt: bool = False,
        stop_on_indicated_failure: bool = True,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send inputs (plural! from a file, line-by-line) on the cli connection.

        Args:
            f: the file to load -- each line will be sent, tralining newlines ignored
            requested_mode: name of the mode to send the input at`
            input_handling: how to handle the input
            retain_input: retain the input in the final "result"
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            stop_on_indicated_failure: stops sending inputs at first indicated failure
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            MultiResult: a MultiResult object representing the operations

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        with open(resolve_file(f), encoding="utf-8", mode="r") as _f:
            inputs = _f.read().splitlines()

        return await self.send_inputs_async(
            inputs=inputs,
            requested_mode=requested_mode,
            input_handling=input_handling,
            retain_input=retain_input,
            retain_trailing_prompt=retain_trailing_prompt,
            stop_on_indicated_failure=stop_on_indicated_failure,
            operation_timeout_ns=operation_timeout_ns,
        )

    def _send_prompted_input(  # noqa: PLR0913
        self,
        *,
        operation_id: OperationIdPointer,
        cancel: CancelPointer,
        input_: c_char_p,
        prompt: c_char_p,
        prompt_pattern: c_char_p,
        response: c_char_p,
        hidden_response: c_bool,
        abort_input: c_char_p,
        requested_mode: c_char_p,
        input_handling: c_char_p,
        retain_trailing_prompt: c_bool,
    ) -> c_uint:
        status = self.ffi_mapping.cli_mapping.send_prompted_input(
            ptr=self._ptr_or_exception(),
            operation_id=operation_id,
            cancel=cancel,
            input_=input_,
            prompt=prompt,
            prompt_pattern=prompt_pattern,
            response=response,
            hidden_response=hidden_response,
            abort_input=abort_input,
            requested_mode=requested_mode,
            input_handling=input_handling,
            retain_trailing_prompt=retain_trailing_prompt,
        )
        if status != 0:
            raise SubmitOperationException("submitting send prompted input operation failed")

        return c_uint(operation_id.contents.value)

    @handle_operation_timeout
    def send_prompted_input(  # noqa: PLR0913
        self,
        input_: str,
        prompt: str,
        prompt_pattern: str,
        response: str,
        *,
        requested_mode: str = "",
        abort_input: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        hidden_response: bool = False,
        retain_trailing_prompt: bool = False,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send a prompted input on the cli connection.

        Args:
            input_: the input to send
            prompt: the prompt to respond to (must set this or prompt_pattern)
            prompt_pattern: the prompt pattern to respond to (must set this or prompt)
            response: the response to send to the prompt
            abort_input: the input to send to abort the "prompted input" operation if an error
                is encountered
            requested_mode: name of the mode to send the input at
            input_handling: how to handle the input
            hidden_response: if the response input will be hidden (like for a password prompt)
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _input = to_c_string(input_)
        _prompt = to_c_string(prompt)
        _prompt_pattern = to_c_string(prompt_pattern)
        _response = to_c_string(response)
        _abort_input = to_c_string(abort_input)
        _requested_mode = to_c_string(requested_mode)
        _input_handling = to_c_string(input_handling)

        operation_id = self._send_prompted_input(
            operation_id=operation_id,
            cancel=cancel,
            input_=_input,
            prompt=_prompt,
            prompt_pattern=_prompt_pattern,
            response=_response,
            abort_input=_abort_input,
            requested_mode=_requested_mode,
            input_handling=_input_handling,
            hidden_response=c_bool(hidden_response),
            retain_trailing_prompt=c_bool(retain_trailing_prompt),
        )

        return self._get_result(operation_id=operation_id)

    @handle_operation_timeout_async
    async def send_prompted_input_async(  # noqa: PLR0913
        self,
        input_: str,
        prompt: str,
        prompt_pattern: str,
        response: str,
        *,
        abort_input: str = "",
        requested_mode: str = "",
        input_handling: InputHandling = InputHandling.FUZZY,
        hidden_response: bool = False,
        retain_trailing_prompt: bool = False,
        operation_timeout_ns: int | None = None,
    ) -> Result:
        """
        Send a prompted input on the cli connection.

        Args:
            input_: the input to send
            prompt: the prompt to respond to (must set this or prompt_pattern)
            prompt_pattern: the prompt pattern to respond to (must set this or prompt)
            response: the response to send to the prompt
            abort_input: the input to send to abort the "prompted input" operation if an error
                is encountered
            requested_mode: name of the mode to send the input at
            input_handling: how to handle the input
            hidden_response: if the response input will be hidden (like for a password prompt)
            retain_trailing_prompt: retain the trailing prompt in the final "result"
            operation_timeout_ns: operation timeout in ns for this operation

        Returns:
            Result: a Result object representing the operation

        Raises:
            NotOpenedException: if the ptr to the cli object is None (via _ptr_or_exception)
            SubmitOperationException: if the operation fails

        """
        # only used in the decorator
        _ = operation_timeout_ns

        operation_id = OperationIdPointer(c_uint(0))
        cancel = CancelPointer(c_bool(False))

        _input = to_c_string(input_)
        _prompt = to_c_string(prompt)
        _prompt_pattern = to_c_string(prompt_pattern)
        _response = to_c_string(response)
        _abort_input = to_c_string(abort_input)
        _requested_mode = to_c_string(requested_mode)
        _input_handling = to_c_string(input_handling)

        operation_id = self._send_prompted_input(
            operation_id=operation_id,
            cancel=cancel,
            input_=_input,
            prompt=_prompt,
            prompt_pattern=_prompt_pattern,
            response=_response,
            abort_input=_abort_input,
            requested_mode=_requested_mode,
            input_handling=_input_handling,
            hidden_response=c_bool(hidden_response),
            retain_trailing_prompt=c_bool(retain_trailing_prompt),
        )

        return await self._get_result_async(operation_id=operation_id)

    @staticmethod
    def ___getwide___() -> None:  # pragma: no cover
        """
        Dumb inside joke easter egg :)

        Args:
            N/A

        Returns:
            None

        Raises:
            N/A

        """
        wide = r"""
KKKXXXXXXXXXXNNNNNNNNNNNNNNNWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW
000000000000KKKKKKKKKKXXXXXXXXXXXXXXXXXNNXXK0Okxdoolllloodxk0KXNNWWNWWWWWWWWWWWWWWWWWWWWWWWWWWWWNNNN
kkkkkkkOOOOOOOOOOO00000000000000000000kdl:,...              ..';coxOKKKKKKKKKKKKXKKXXKKKXXXXXKKKK000
kkkkkkkOOOOOOOOOOOO000000000000000Od:,.                            .,cdOKKKKKKKKKKKK0000OOOOOOOOOOOO
kkkkkkkkOOOOOOOOOOO0000000000000kc'                                    .:d0KKKKKKKKK0KKOkOOOOOOOOOO0
kkkkkkkkOOOOOOOOOOOO00000000000o'                                         ,o0KKKKKKKKKKOkOOOOOOOOO00
kkkkkkkkOOOOOOOOOOOOO000000000o.                                            ;kKKKKKKKKKOkOOOOOOOOO00
OOOOOOOOOO0000000000000000K0Kk'                                              'xKKKKKKKKOkOOOOOOOOO00
KKKKKKKKKXXXXXXXXXXXXXXNNNNNNd.                                               cXNNNNNNNK0000O00O0000
KKKKKKKKKXXXXXXXXXXXXNNNNNNNXl        ...............                         :XWWWWWWWX000000000000
KKKKKKKKKXXXXXXXXXXXXXXNNNNNXc     ...''',,,,,,;;,,,,,,'''......             .xWWWWWWWWX000000000000
KKKKKKKKKKKXXXXXXXXXXXXXNNNNK;    ...',,,,;;;;;;;:::::::;;;;;;,,'.          .oNWWWWWWWNK000000OOOO00
KKKKKKKKKKKKXXXXXXXXXXXXXXXN0,  ...'',,,;;;;;;:::::::::::::::;;;;,'.       .dNWWWWWWWWNK0000OOOOOOOO
0000KKKKKKKKKKKKKXXXXXXXXXXN0, ..'',,,,;;;;;;:::::::::::::::::;;;;,,..    ;ONNNNNWWWWWNK00OOOOOOOOOO
kkkkkkOOOOOOOOOOOOOOOOOOO000k; ..,,,,,,'',,;;::::::::::::::::;;;;;;,'.  .lOKKKKXXKXXKK0OOOOOOOOOOOOO
xxxkkkkkkkkkkkkkkkkkkOOOOkdll;..',,,,,,,''...';::ccccc:::::::::;;;;;,...o0000000000000OkkOOOkkOOOOOO
xxxxxxkkkkkkkkkkkkkkkkkkOd:;;,..,;;;;;;;;;;,'',,;:ccccccccc:::;;;;;;,..cO0000000000000Oxkkkkkkkkkkkk
xxxxxxxxkkkkkkkkkkkkkkkkkl:;;,'';;;;;,'''''',,,,,;::ccc::;,,'.'''',;,,lO00000000000000kxkkkkkkkkkkkk
xxxxxxxxkkkkkkkkkkkkkkkkko::;'';;;;;;,''....,'',,,,;:c:;,,'''',,;;;;,:x00000000000000Okxkkkkkkkkkkkk
xxxxxxxxxxkkkkkkkkkkkkkkkxl;,,;;;;:::;;;,,,,,,,,,,,,:c:;,'....''',;;,;cxO000000000000Okxkkkkkkkkkkkk
kkkkOOOOOOOOOOOOOO00000000x:;;;;;:::c::::::;;;;;;;;;:c:;,,,,'',,',;:::lOKKKKKKXXXXXXKKOkkkkkkkkkkkkk
000000000000000KKKKKKKKKKK0dc;,;;:::ccccccc::::;;;;;:cc:;;;;:::::::::lOXXXXXNNNNNNNNXX0Okkkkkkkkkkkk
OO00000000000000000KKKKKKKK0d::;;;::ccccccccc:;;;;;;;:c:;::ccccccc::cOXXXXXXXXXNNNNNXX0kkkkkkkkkkkkk
OOO00000000000000000000KKKKKOxxc;;;::ccccccc:;;;;;;;:ccc:::cccllcc;:kKXXXXXXXXXXXXXXXKOkkkkkkkkkkkkk
OOOOO00000000000000000000KKK0kdl;;;;;:ccccc::;,,,,;;:clc:::cclllcc:oKXXXXXXXXXXXXXXXXKOkkkkkkkkkkkkk
OOOOOOO0000000000000000Okxdlc;,,;;::;;::cc::;;,,,,,;:::;;:cccccc::clxkO00KKKKKKKKKXKK0kkkkkkkkkkkxkk
kkkkkkkkkkkkkkkkkkkxdoc:,''.....,;:::;;;::;;;;;;;;;;;;;;;:ccc:::;,',;;:clodxkOOOOOOOOkxxxxxxxxxxxxxx
ddddddddddddddoolc;,'''..........,;;:;;;::;,,,,,;;;;;::::::c:::;'.',,;;;;;::clodxkkkkxdxxxxxxxxxxxxx
dddddddoolc::;,'''.......      ..',;;;;;;;;,'........',;::::::;;,,;;;;;;;;:::::ccloddddxxxxxxxxxxxxx
dollc:;,,''.........         ..'''',,,,;;;;;,'''.....'',::::;,,;;;::::;;,,;;;;;;;;;::cldxxxxxxdxxdxx
l;'''.''......             ..'',,''',,,,;;;::;;,,,,,,;;::;;'.....',;;,,''',,,,,,'',,,',:odxddddddddd
.............             .'',,,,,''',,,;;;;::::;::::::;;;........'''''''..'.....,,'...';cdddddddddd
. .......                .',,,,,;,,'',,,,;;;::::::::::;;cc. .....''...'''.......','......':odxdddddd
   ...                  .',,;;;;;;,'',;;,,,;;;::::::::;cxo....................''''.......'';lddddddd
    ..                  .,;,;;;;;;,,,',;;;,,,,;;;;;;;;:dKO:..................''''.. .......',cdddddd
                         ,:;;;;;,,,,;,,;::;,,,,,;::::::dK0c..................'''..  ........',codddd
                         .;:;;;;;,,;;;,,;:;;:;,,;:::::clc,...   ...........'''.... ....  .....':oddd
                          .',;;;;;;;;;,,;:;;;;,;::::::;'......       ......'.........   .....'',cood
                            ..,;;;;;;;;;;;:;;;;:::::;'.    .         ..............       ...''',:od
                              ..',;;;;:;;;:::::::,,'.              ...............        ....''.':o
                                 ...',,;;,,;,,'..                 ...............        ..  .....'c
               __              _     __....                      ................     ....   ......'
   ____ ____  / /_   _      __(_)___/ /__                    ..............   ..    ...     .......
  / __ `/ _ \/ __/  | | /| / / / __  / _ \                 ................    .             ......
 / /_/ /  __/ /_    | |/ |/ / / /_/ /  __/                .................                  ......
 \__, /\___/\__/    |__/|__/_/\__,_/\___/                  ...............                   ......
/____/                                                     ...............  ..             ........
"""
        print(wide)



"""scrapli.ffi"""

import importlib.resources
import os
import sys
from logging import getLogger
from pathlib import Path

from scrapli.exceptions import LibScrapliException

logger = getLogger(__name__)

LIBSCRAPLI_VERSION = "0.0.1-alpha.10"
LIBSCRAPLI_PATH_OVERRIDE_ENV = "LIBSCRAPLI_PATH"
LIBSCRAPLI_CACHE_PATH_OVERRIDE_ENV = "LIBSCRAPLI_CACHE_PATH"
XDG_CACHE_HOME_ENV = "XDG_CACHE_HOME"


def get_libscrapli_shared_object_filename(version: str = LIBSCRAPLI_VERSION) -> str:
    """
    Returns the name of the libscrapli shared object for the given version/platform.

    Args:
        version: the libscrapli version

    Returns:
        str: filename of the shared object

    Raises:
        LibScrapliException: if unsupported platform

    """
    if sys.platform == "linux":
        lib_filename = f"libscrapli.so.{version}"
    elif sys.platform == "darwin":
        lib_filename = f"libscrapli.{version}.dylib"
    else:
        raise LibScrapliException("unsupported platform")

    return lib_filename


def get_libscrapli_path() -> str:
    """
    Returns the file path to the libscrapli shared library.

    Attempts to load from override paths or from installed path in scrapli itself -- this would be
    either the shared object(s) from source (i.e. cloning the repo) or from the installation either
    via a wheel or sdist.

    Args:
        N/A

    Returns:
        str: libscrapli shared library path

    Raises:
        LibScrapliException: if libscrapli is not found at override path or default installation
            path

    """
    override_path = os.environ.get(LIBSCRAPLI_PATH_OVERRIDE_ENV)
    if override_path is not None:
        logger.debug("using libscrapli path override '%s'", override_path)

        return override_path

    source_lib_dir = importlib.resources.files("scrapli.lib")
    source_lib_filename = f"{source_lib_dir}/{get_libscrapli_shared_object_filename()}"

    logger.debug("loading libscrapli from scrapli installation '%s'", source_lib_filename)

    if Path(source_lib_filename).exists():
        return source_lib_filename

    raise LibScrapliException("libscrapli not available")



"""scrapli.auth"""

from ctypes import c_char_p
from dataclasses import dataclass, field

from scrapli.exceptions import OptionsException
from scrapli.ffi_mapping import LibScrapliMapping
from scrapli.ffi_types import (
    DriverPointer,
    to_c_string,
)


@dataclass
class LookupKeyValue:
    """
    Options a "lookup" key/value pair.

    Used in conjuection with platform definition templating like `__lookup::enable` where "enable"
    is the key to "lookup" in the list of lookup key/values.

    Args:
        key: the name of the lookup key
        value: the value of the lookup key

    Returns:
        None

    Raises:
        N/A

    """

    key: str
    value: str

    _key: c_char_p | None = field(init=False, default=None, repr=False)
    _value: c_char_p | None = field(init=False, default=None, repr=False)

    def _get_c_strings(self) -> tuple[c_char_p, c_char_p]:
        self._key = to_c_string(self.key)
        self._value = to_c_string(self.value)

        return self._key, self._value

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(key='{self.key}', value='REDACTED')"


@dataclass
class Options:
    """
    Options holds auth related options to pass to the ffi layer.

    All arguments are optional, though you will almost certainly need to provide *some* -- for
    example a username and password or key.

    Args:
        username: the username to use for the connection
        password: the password to use for the connection
        private_key_path: filepath to the ssh private key to use
        private_key_passphrase: the private key passphrase if set
        lookups: a list of key/values that can be "looked up" from a connection -- used in
            conjunction with platform definition templating like `__lookup::enable` where "enable"
            is the key to "lookup" in the list of lookup key/values.
        in_session_auth_bypass: whether to bypass attempting in session authentication -- only
             applicable to "bin" and "telnet" transports.
        username_pattern: the regex pattern to use to look for a username prompt
        password_pattern: the regex pattern to use to look for a password prompt
        private_key_passphrase_pattern: the regex pattern to use to look for a passphrase prompt

    Returns:
        None

    Raises:
        N/A

    """

    username: str | None = None
    password: str | None = None
    private_key_path: str | None = None
    private_key_passphrase: str | None = None
    lookups: list[LookupKeyValue] | None = None
    in_session_auth_bypass: bool | None = None
    username_pattern: str | None = None
    password_pattern: str | None = None
    private_key_passphrase_pattern: str | None = None

    _username: c_char_p | None = field(init=False, default=None, repr=False)
    _password: c_char_p | None = field(init=False, default=None, repr=False)
    _private_key_path: c_char_p | None = field(init=False, default=None, repr=False)
    _private_key_passphrase: c_char_p | None = field(init=False, default=None, repr=False)
    _username_pattern: c_char_p | None = field(init=False, default=None, repr=False)
    _password_pattern: c_char_p | None = field(init=False, default=None, repr=False)
    _private_key_passphrase_pattern: c_char_p | None = field(init=False, default=None, repr=False)

    def apply(  # noqa: C901, PLR0912
        self, ffi_mapping: LibScrapliMapping, ptr: DriverPointer
    ) -> None:
        """
        Applies the options to the given driver pointer.

        Should not be called directly/by users.

        Args:
            ffi_mapping: the handle to the ffi mapping singleton
            ptr: the pointer to the underlying cli or netconf object

        Returns:
            None

        Raises:
            OptionsException: if any option apply returns a non-zero return code.

        """
        if self.username is not None:
            self._username = to_c_string(self.username)

            status = ffi_mapping.options_mapping.auth.set_username(ptr, self._username)
            if status != 0:
                raise OptionsException("failed to set auth username")

        if self.password is not None:
            self._password = to_c_string(self.password)

            status = ffi_mapping.options_mapping.auth.set_password(ptr, self._password)
            if status != 0:
                raise OptionsException("failed to set auth password")

        if self.private_key_path is not None:
            self._private_key_path = to_c_string(self.private_key_path)

            status = ffi_mapping.options_mapping.auth.set_private_key_path(
                ptr, self._private_key_path
            )
            if status != 0:
                raise OptionsException("failed to set auth private key path")

        if self.private_key_passphrase is not None:
            self._private_key_passphrase = to_c_string(self.private_key_passphrase)

            status = ffi_mapping.options_mapping.auth.set_private_key_passphrase(
                ptr, self._private_key_passphrase
            )
            if status != 0:
                raise OptionsException("failed to set auth private key passphrase")

        if self.lookups is not None:
            for lookup in self.lookups:
                status = ffi_mapping.options_mapping.auth.set_lookup_key_value(
                    ptr, *lookup._get_c_strings()
                )
                if status != 0:
                    raise OptionsException("failed to set auth lookup key/value")

        if self.in_session_auth_bypass is not None:
            status = ffi_mapping.options_mapping.auth.set_in_session_auth_bypass(ptr)
            if status != 0:
                raise OptionsException("failed to set session in session auth bypass")

        if self.username_pattern is not None:
            self._username_pattern = to_c_string(self.username_pattern)

            status = ffi_mapping.options_mapping.auth.set_username_pattern(
                ptr, self._username_pattern
            )
            if status != 0:
                raise OptionsException("failed to set auth username pattern")

        if self.password_pattern is not None:
            self._password_pattern = to_c_string(self.password_pattern)

            status = ffi_mapping.options_mapping.auth.set_password_pattern(
                ptr, self._password_pattern
            )
            if status != 0:
                raise OptionsException("failed to set auth password pattern")

        if self.private_key_passphrase_pattern is not None:
            self._private_key_passphrase_pattern = to_c_string(self.private_key_passphrase_pattern)

            status = ffi_mapping.options_mapping.auth.set_private_key_passphrase_pattern(
                ptr, self._private_key_passphrase_pattern
            )
            if status != 0:
                raise OptionsException("failed to set auth private key passphrase pattern")

    def __repr__(self) -> str:
        """
        Magic repr method for Options object

        Args:
            N/A

        Returns:
            str: repr for Options object

        Raises:
            N/A

        """
        return (
            # it will probably be "canonical" to import Options as AuthOptions, so we'll make
            # the repr do that too
            f"Auth{self.__class__.__name__}("
            f"username={self.username!r}, "
            "password=REDACTED, "
            f"private_key_path={self.private_key_path!r} "
            f"private_key_passphrase={self.private_key_passphrase!r} "
            f"lookups={self.lookups!r}) "
            f"in_session_auth_bypass={self.in_session_auth_bypass!r}) "
            f"username_pattern={self.username_pattern!r}) "
            f"password_pattern={self.password_pattern!r}) "
            f"private_key_passphrase_pattern={self.private_key_passphrase_pattern!r}) "
        )



"""scrapli.result"""

from typing import Any, TextIO

from scrapli.cli_parse import genie_parse, textfsm_get_template, textfsm_parse
from scrapli.exceptions import ParsingException

OPERATION_DELIMITER = "__libscrapli__"


class Result:
    """
    Result represents a set of results from some Cli operation(s).

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(  # noqa: PLR0913
        self,
        *,
        host: str,
        port: int,
        inputs: str,
        start_time: int,
        splits: list[int],
        results_raw: bytes,
        results: str,
        results_failed_indicator: str,
        textfsm_platform: str,
        genie_platform: str,
    ) -> None:
        self.host = host
        self.port = port
        self.inputs = inputs.split(OPERATION_DELIMITER)
        self.start_time = start_time
        self.splits = splits
        self.results_raw = results_raw.split(OPERATION_DELIMITER.encode())
        self.results = results.split(OPERATION_DELIMITER)
        self.results_failed_indicator = results_failed_indicator
        self.textfsm_platform = textfsm_platform
        self.genie_platform = genie_platform

    def extend(self, result: "Result") -> None:
        """
        Extends this Result object with another Result object.

        Args:
            result: the result object with which to extend this result object

        Returns:
            N/A

        Raises:
            N/A

        """
        self.inputs.extend(result.inputs)
        self.results_raw.extend(result.results_raw)
        self.results.extend(result.results)
        self.splits.extend(result.splits)

    @property
    def failed(self) -> bool:
        """
        Returns True if any failed indicators were seen, otherwise False.

        Args:
            N/A

        Returns:
            bool: True for failed, otherwise False

        Raises:
            N/A

        """
        return bool(self.results_failed_indicator)

    @property
    def end_time(self) -> int:
        """
        Returns the end time of the operations in unix nano.

        Args:
            N/A

        Returns:
            int: end time in unix nano

        Raises:
            N/A

        """
        if not self.splits:
            # if we had no splits it was a "noop" type op (like enter mode when
            # you are already in the requested mode), so we'll lie and say it
            # was a 1ns op time
            return self.start_time + 1

        return self.splits[-1]

    @property
    def elapsed_time_seconds(self) -> float:
        """
        Returns the number of seconds the operation took.

        Args:
            N/A

        Returns:
            float: duration in seconds

        Raises:
            N/A

        """
        return (self.end_time - self.start_time) / 1_000_000_000

    @property
    def result(self) -> str:
        """
        Returns the results joined on newline chars. Note this does *not* include inputs sent.

        Args:
            N/A

        Returns:
            str: joined results

        Raises:
            N/A

        """
        return "\n".join(self.results)

    def textfsm_parse(
        self,
        index: int = 0,
        template: str | TextIO | None = None,
        to_dict: bool = True,
    ) -> list[Any] | dict[str, Any]:
        """
        Parse results with textfsm, always return structured data

        Returns an empty list if parsing fails!

        Args:
            index: the index of the result to parse, assumes first/zeroith if not provided
            template: string path to textfsm template or opened textfsm template file
            to_dict: convert textfsm output from list of lists to list of dicts -- basically create
                dict from header and row data so it is easier to read/parse the output

        Returns:
            structured_result: empty list or parsed data from textfsm

        Raises:
            N/A

        """
        if template is None:
            template = textfsm_get_template(
                platform=self.textfsm_platform, command=self.inputs[index]
            )

        if template is None:
            raise ParsingException("no template provided or available for input")

        return textfsm_parse(template=template, output=self.result, to_dict=to_dict)

    def genie_parse(
        self,
        index: int = 0,
    ) -> dict[str, Any] | list[Any]:
        """
        Parse results with genie, always return structured data

        Returns an empty list if parsing fails!

        Args:
            index: the index of the result to parse, assumes first/zeroith if not provided

        Returns:
            structured_result: empty list or parsed data from genie

        Raises:
            N/A

        """
        return genie_parse(self.genie_platform, self.inputs[index], self.results[index])



"""scrapli.ffi_mapping_options"""

from collections.abc import Callable
from ctypes import (
    CDLL,
    c_char_p,
    c_int,
    c_uint8,
    c_uint64,
)

from scrapli.ffi_types import (
    DriverPointer,
)


class LibScrapliNetconfOptionsMapping:
    """
    Mapping to libscrapli netconf option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._set_error_tag: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_netconf_error_tag
        )
        lib.ls_option_netconf_error_tag.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_netconf_error_tag.restype = c_uint8

        self._set_preferred_version: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_netconf_preferred_version
        )
        lib.ls_option_netconf_preferred_version.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_netconf_preferred_version.restype = c_uint8

        self._set_message_poll_interva_ns: Callable[[DriverPointer, c_int], int] = (
            lib.ls_option_netconf_message_poll_interval
        )
        lib.ls_option_netconf_message_poll_interval.argtypes = [
            DriverPointer,
            c_int,
        ]
        lib.ls_option_netconf_message_poll_interval.restype = c_uint8

    def set_error_tag(self, ptr: DriverPointer, error_tag: c_char_p) -> int:
        """
        Set the error tag substring.

        Should not be used/called directly.

        Args:
            ptr: ptr to the netconf object
            error_tag: error tag substring

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_error_tag(ptr, error_tag)

    def set_preferred_version(self, ptr: DriverPointer, version: c_char_p) -> int:
        """
        Set the preferred netconf version.

        Should not be used/called directly.

        Args:
            ptr: ptr to the netconf object
            version: version string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_preferred_version(ptr, version)

    def set_message_poll_interva_ns(self, ptr: DriverPointer, interval: c_int) -> int:
        """
        Set the netconf message poll interval.

        Should not be used/called directly.

        Args:
            ptr: ptr to the netconf object
            interval: interval in ns

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_message_poll_interva_ns(ptr, interval)


class LibScrapliSessionOptionsMapping:
    """
    Mapping to libscrapli session option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._set_read_size: Callable[
            [
                DriverPointer,
                c_uint64,
            ],
            int,
        ] = lib.ls_option_session_read_size
        lib.ls_option_session_read_size.argtypes = [
            DriverPointer,
            c_uint64,
        ]
        lib.ls_option_session_read_size.restype = c_uint8

        self._set_return_char: Callable[
            [
                DriverPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_option_session_return_char
        lib.ls_option_session_return_char.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_session_return_char.restype = c_uint8

        self._set_operation_timeout_ns: Callable[
            [
                DriverPointer,
                c_uint64,
            ],
            int,
        ] = lib.ls_option_session_operation_timeout_ns
        lib.ls_option_session_operation_timeout_ns.argtypes = [
            DriverPointer,
            c_uint64,
        ]
        lib.ls_option_session_operation_timeout_ns.restype = c_uint8

        self._set_operation_max_search_depth: Callable[
            [
                DriverPointer,
                c_uint64,
            ],
            int,
        ] = lib.ls_option_session_operation_max_search_depth
        lib.ls_option_session_operation_max_search_depth.argtypes = [
            DriverPointer,
            c_uint64,
        ]
        lib.ls_option_session_operation_max_search_depth.restype = c_uint8

        self._set_recorder_path: Callable[
            [
                DriverPointer,
                c_char_p,
            ],
            int,
        ] = lib.ls_option_session_record_destination
        lib.ls_option_session_record_destination.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_session_record_destination.restype = c_uint8

    def set_read_size(self, ptr: DriverPointer, read_size: c_uint64) -> int:
        """
        Set the session read size.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            read_size: read size

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_read_size(ptr, read_size)

    def set_return_char(self, ptr: DriverPointer, return_char: c_char_p) -> int:
        r"""
        Set the session return char

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            return_char: return char to use in the session (usually \n, could be \r\n)

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_return_char(ptr, return_char)

    def set_operation_timeout_ns(
        self, ptr: DriverPointer, set_operation_timeout_ns: c_uint64
    ) -> int:
        """
        Set the session operation timeout in ns.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            set_operation_timeout_ns: operation timeout in ns

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_operation_timeout_ns(ptr, set_operation_timeout_ns)

    def set_operation_max_search_depth(
        self, ptr: DriverPointer, set_operation_max_search_depth: c_uint64
    ) -> int:
        """
        Set the session maximum prompt search depth

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            set_operation_max_search_depth: maximum prompt search depth (in count of chars)

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_operation_max_search_depth(ptr, set_operation_max_search_depth)

    def set_recorder_path(self, ptr: DriverPointer, set_recorder_path: c_char_p) -> int:
        """
        Set the session recorder output path

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            set_recorder_path: file path to write session recorder output to

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_recorder_path(ptr, set_recorder_path)


class LibScrapliAuthOptionsMapping:
    """
    Mapping to libscrapli auth option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._set_username: Callable[[DriverPointer, c_char_p], int] = lib.ls_option_auth_username
        lib.ls_option_auth_username.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_username.restype = c_uint8

        self._set_password: Callable[[DriverPointer, c_char_p], int] = lib.ls_option_auth_password
        lib.ls_option_auth_password.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_password.restype = c_uint8

        self._set_private_key_path: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_auth_private_key_path
        )
        lib.ls_option_auth_private_key_path.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_private_key_path.restype = c_uint8

        self._set_private_key_passphrase: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_auth_private_key_passphrase
        )
        lib.ls_option_auth_private_key_passphrase.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_private_key_passphrase.restype = c_uint8

        self._set_lookup_key_value: Callable[[DriverPointer, c_char_p, c_char_p], int] = (
            lib.ls_option_auth_set_lookup_key_value
        )
        lib.ls_option_auth_set_lookup_key_value.argtypes = [
            DriverPointer,
            c_char_p,
            c_char_p,
        ]
        lib.ls_option_auth_set_lookup_key_value.restype = c_uint8

        self._set_in_session_auth_bypass: Callable[[DriverPointer], int] = (
            lib.ls_option_auth_in_session_auth_bypass
        )
        lib.ls_option_auth_in_session_auth_bypass.argtypes = [
            DriverPointer,
        ]
        lib.ls_option_auth_in_session_auth_bypass.restype = c_uint8

        self._set_username_pattern: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_auth_username_pattern
        )
        lib.ls_option_auth_username_pattern.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_username_pattern.restype = c_uint8

        self._set_password_pattern: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_auth_password_pattern
        )
        lib.ls_option_auth_password_pattern.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_password_pattern.restype = c_uint8

        self._set_private_key_passphrase_pattern: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_auth_private_key_passphrase_pattern
        )
        lib.ls_option_auth_private_key_passphrase_pattern.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_auth_private_key_passphrase_pattern.restype = c_uint8

    def set_username(self, ptr: DriverPointer, username: c_char_p) -> int:
        """
        Set the username.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            username: username string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_username(ptr, username)

    def set_password(self, ptr: DriverPointer, password: c_char_p) -> int:
        """
        Set the password.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            password: password string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_password(ptr, password)

    def set_private_key_path(self, ptr: DriverPointer, private_key_path: c_char_p) -> int:
        """
        Set the private key path.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            private_key_path: private key path string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_private_key_path(ptr, private_key_path)

    def set_private_key_passphrase(
        self, ptr: DriverPointer, private_key_passphrase: c_char_p
    ) -> int:
        """
        Set the private key passphrase.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            private_key_passphrase: private key passphrase string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_private_key_passphrase(ptr, private_key_passphrase)

    def set_lookup_key_value(
        self,
        ptr: DriverPointer,
        key: c_char_p,
        value: c_char_p,
    ) -> int:
        """
        Set a lookup key/value pair.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            key: the name of the key
            value: the value of the key

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_lookup_key_value(ptr, key, value)

    def set_in_session_auth_bypass(
        self,
        ptr: DriverPointer,
    ) -> int:
        """
        Enable the in session auth bypass.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_in_session_auth_bypass(
            ptr,
        )

    def set_username_pattern(self, ptr: DriverPointer, pattern: c_char_p) -> int:
        """
        Set the username pattern.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            pattern: username pattern string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_username_pattern(ptr, pattern)

    def set_password_pattern(self, ptr: DriverPointer, pattern: c_char_p) -> int:
        """
        Set the password pattern.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            pattern: password pattern string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_password_pattern(ptr, pattern)

    def set_private_key_passphrase_pattern(self, ptr: DriverPointer, pattern: c_char_p) -> int:
        """
        Set the private key passphrase pattern.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            pattern: private key passphrase pattern string

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_private_key_passphrase_pattern(ptr, pattern)


class LibScrapliTransportBinOptionsMapping:
    """
    Mapping to libscrapli bin (default) transport option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._set_bin: Callable[[DriverPointer, c_char_p], int] = lib.ls_option_transport_bin_bin
        lib.ls_option_transport_bin_bin.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_bin_bin.restype = c_uint8

        self._set_extra_open_args: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_transport_bin_extra_open_args
        )
        lib.ls_option_transport_bin_extra_open_args.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_bin_extra_open_args.restype = c_uint8

        self._set_override_open_args: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_transport_bin_override_open_args
        )
        lib.ls_option_transport_bin_override_open_args.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_bin_override_open_args.restype = c_uint8

        self._set_ssh_config_path: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_transport_bin_ssh_config_path
        )
        lib.ls_option_transport_bin_ssh_config_path.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_bin_ssh_config_path.restype = c_uint8

        self._set_known_hosts_path: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_transport_bin_known_hosts_path
        )
        lib.ls_option_transport_bin_known_hosts_path.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_bin_known_hosts_path.restype = c_uint8

        self._set_enable_strict_key: Callable[[DriverPointer], int] = (
            lib.ls_option_transport_bin_enable_strict_key
        )
        lib.ls_option_transport_bin_enable_strict_key.argtypes = [
            DriverPointer,
        ]
        lib.ls_option_transport_bin_enable_strict_key.restype = c_uint8

        self._set_term_height: Callable[[DriverPointer, c_int], int] = (
            lib.ls_option_transport_bin_term_height
        )
        lib.ls_option_transport_bin_term_height.argtypes = [
            DriverPointer,
            c_int,
        ]
        lib.ls_option_transport_bin_term_height.restype = c_uint8

        self._set_term_width: Callable[[DriverPointer, c_int], int] = (
            lib.ls_option_transport_bin_term_width
        )
        lib.ls_option_transport_bin_term_width.argtypes = [
            DriverPointer,
            c_int,
        ]
        lib.ls_option_transport_bin_term_width.restype = c_uint8

    def set_bin(self, ptr: DriverPointer, bin_: c_char_p) -> int:
        """
        Set bin transport binary file to exec.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            bin_: path to the bin to use

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_bin(ptr, bin_)

    def set_extra_open_args(
        self,
        ptr: DriverPointer,
        extra_open_args: c_char_p,
    ) -> int:
        """
        Set bin transport extra open args.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            extra_open_args: extra open args

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_extra_open_args(
            ptr,
            extra_open_args,
        )

    def set_override_open_args(
        self,
        ptr: DriverPointer,
        override_open_args: c_char_p,
    ) -> int:
        """
        Set bin transport extra open args.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            override_open_args: override open args

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_override_open_args(
            ptr,
            override_open_args,
        )

    def set_ssh_config_path(self, ptr: DriverPointer, path: c_char_p) -> int:
        """
        Set bin transport ssh config file path.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            path: the path to the ssh config file

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_ssh_config_path(ptr, path)

    def set_known_hosts_path(self, ptr: DriverPointer, path: c_char_p) -> int:
        """
        Set bin transport known hosts file path.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            path: the path to the known hosts file

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_known_hosts_path(ptr, path)

    def set_enable_strict_key(
        self,
        ptr: DriverPointer,
    ) -> int:
        """
        Set bin transport strict key checking

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_enable_strict_key(
            ptr,
        )

    def set_term_height(self, ptr: DriverPointer, height: c_int) -> int:
        """
        Set bin transport terminal height

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            height: size of terminal height

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_term_height(ptr, height)

    def set_term_width(self, ptr: DriverPointer, width: c_int) -> int:
        """
        Set bin transport terminal width

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            width: size of terminal width

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_term_width(ptr, width)


class LibScrapliTransportSsh2OptionsMapping:
    """
    Mapping to libscrapli ssh2 transport option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._set_known_hosts_path: Callable[[DriverPointer, c_char_p], int] = (
            lib.ls_option_transport_ssh2_known_hosts_path
        )
        lib.ls_option_transport_ssh2_known_hosts_path.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_ssh2_known_hosts_path.restype = c_uint8

        self._set_libssh2_trace: Callable[[DriverPointer], int] = (
            lib.ls_option_transport_ssh2_libssh2trace
        )
        lib.ls_option_transport_ssh2_libssh2trace.argtypes = [
            DriverPointer,
        ]
        lib.ls_option_transport_ssh2_libssh2trace.restype = c_uint8

    def set_known_hosts_path(self, ptr: DriverPointer, path: c_char_p) -> int:
        """
        Set ssh2 transport known hosts file path.

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            path: the path to the known hosts file

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_known_hosts_path(ptr, path)

    def set_libssh2_trace(self, ptr: DriverPointer) -> int:
        """
        Set ssh2 transport libssh2 trace

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_libssh2_trace(ptr)


class LibScrapliTransportTelnetOptionsMapping:
    """
    Mapping to libscrapli telnet transport option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        _ = lib


class LibScrapliTransportTestOptionsMapping:
    """
    Mapping to libscrapli test transport option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self._set_f: Callable[[DriverPointer, c_char_p], int] = lib.ls_option_transport_test_f
        lib.ls_option_transport_test_f.argtypes = [
            DriverPointer,
            c_char_p,
        ]
        lib.ls_option_transport_test_f.restype = c_int

    def set_f(self, ptr: DriverPointer, f: c_char_p) -> int:
        """
        Set test transport f

        Should not be used/called directly.

        Args:
            ptr: ptr to the cli/netconf object
            f: file to read/load

        Returns:
            int: return code, non-zero value indicates an error. technically a c_uint8 converted by
                ctypes.

        Raises:
            N/A

        """
        return self._set_f(ptr, f)


class LibScrapliOptionsMapping:
    """
    Mapping to libscrapli option setter exported functions.

    Should not be used/called directly.

    Args:
        N/A

    Returns:
        None

    Raises:
        N/A

    """

    def __init__(self, lib: CDLL) -> None:
        self.netconf = LibScrapliNetconfOptionsMapping(lib)
        self.session = LibScrapliSessionOptionsMapping(lib)
        self.auth = LibScrapliAuthOptionsMapping(lib)
        self.transport_bin = LibScrapliTransportBinOptionsMapping(lib)
        self.transport_ssh2 = LibScrapliTransportSsh2OptionsMapping(lib)
        self.transport_telnet = LibScrapliTransportTelnetOptionsMapping(lib)
        self.transport_test = LibScrapliTransportTestOptionsMapping(lib)



"""setup"""

import os
import platform
import re
import shutil
import subprocess
import sys
import tempfile
from http import HTTPStatus
from pathlib import Path
from urllib.request import urlopen

from setuptools import setup
from setuptools.command.bdist_wheel import bdist_wheel
from setuptools.command.editable_wheel import editable_wheel
from setuptools.command.sdist import sdist

LIBSCRAPLI_VERSION = "0.0.1-alpha.10"
LIBSCRAPLI_REPO = "https://github.com/scrapli/libscrapli"
LIBSCRAPLI_BUILD_PATH_ENV = "LIBSCRAPLI_BUILD_PATH"
LIBSCRPALI_ZIG_TRIPLE_ENV = "LIBSCRAPLI_ZIG_TRIPLE"

# if set to anything, always build libscrapli, as in: dont just slurp down the artifacts
# from a release
LIBSCRAPLI_ALWAYS_BUILD_ENV = "LIBSCRAPLI_ALWAYS_BUILD"
LIBSCRAPLI_ALWAYS_BUILD = (
    True if os.getenv(LIBSCRAPLI_ALWAYS_BUILD_ENV, None) is not None else False
)

WHEEL_TARGETS = {
    # zig-triple <-> wheel
    "x86_64-macos": "macosx_11_0_x86_64",
    "aarch64-macos": "macosx_11_0_arm64",
    "x86_64-linux-gnu": "manylinux2010_x86_64",
    "x86_64-linux-musl": "musllinux_1_1_x86_64",
    "aarch64-linux-gnu": "manylinux2014_aarch64",
    "aarch64-linux-musl": "musllinux_1_1_aarch64",
}


class Libscrapli:
    """Dumb container for setup-related helpers"""

    _libscrapli_tag: str | None = None
    _libscrapli_version_is_release: bool | None = None

    @property
    def libscrapli_tag(self) -> str:
        """
        Return the libscrapli version/tag

        Args:
            N/A

        Returns:
            str: the parsed libscrapli verison/tag

        Raises:
            N/A

        """
        if self._libscrapli_tag is not None:
            return self._libscrapli_tag

        self._libscrapli_tag = (
            f"v{LIBSCRAPLI_VERSION}"
            if not re.fullmatch(r"[0-9a-f]{7,40}", LIBSCRAPLI_VERSION)
            else LIBSCRAPLI_VERSION
        )

        return self._libscrapli_tag

    @property
    def libscrapli_is_release_tag(self) -> bool:
        """
        Return true if the target libscrapli version is a released version, othewrise false

        Args:
            N/A

        Returns:
            bool: true if a released (semver) version

        Raises:
            N/A

        """
        if self._libscrapli_version_is_release is not None:
            return self._libscrapli_version_is_release

        if self.libscrapli_tag == "" or re.fullmatch(r"[0-9a-f]{7,40}", self.libscrapli_tag):
            self._libscrapli_version_is_release = False
        else:
            self._libscrapli_version_is_release = True

        return self._libscrapli_version_is_release

    @staticmethod
    def _get_zig_style_arch() -> str:
        p = platform.machine()

        if p in {"amd64", "x86_64"}:
            return "x86_64"

        if p in {"arm64", "aarch64"}:
            return "aarch64"

        raise NotImplementedError(f"unsupported platform '{p}'")

    @staticmethod
    def _get_zig_style_platform() -> str:
        if sys.platform == "darwin":
            return "macos"

        if sys.platform == "linux":
            if os.path.exists("/lib/libc.musl-x86_64.so.1"):
                return "linux-musl"
            return "linux-gnu"

        raise NotImplementedError(f"unsupported platform '{sys.platform}'")

    @staticmethod
    def _get_libscrapli_build_path() -> Path:
        if libscrapli_build_path := os.environ.get(LIBSCRAPLI_BUILD_PATH_ENV):
            return Path(libscrapli_build_path)

        return Path(tempfile.mkdtemp())

    @staticmethod
    def _clean_libscrapli_build_path(libscrapli_build_path: Path) -> None:
        if LIBSCRAPLI_BUILD_PATH_ENV in os.environ:
            # build path env set, likely in ci, dont clean up as we may need it for
            # building other wheels. ci will/should clean it
            return

        shutil.rmtree(libscrapli_build_path)

    @staticmethod
    def _get_libscrapli_output_shared_object_filename() -> str:
        p = sys.platform

        zig_triple = os.environ.get(LIBSCRPALI_ZIG_TRIPLE_ENV, None)
        if zig_triple is not None:
            # probably building all the wheels, we dont actually want the shared object
            # for our current system, we want it for the target triple
            if "linux" in zig_triple:
                p = "linux"
            else:
                p = "darwin"

        if p == "linux":
            lib_filename = f"libscrapli.so.{LIBSCRAPLI_VERSION}"
        elif p == "darwin":
            lib_filename = f"libscrapli.{LIBSCRAPLI_VERSION}.dylib"
        else:
            raise NotImplementedError("unsupported platform")

        return lib_filename

    def _get_libscrapli_built_shared_object_filename(self, build_path: str) -> str:
        if self.libscrapli_is_release_tag:
            return str(Path(build_path) / self._get_libscrapli_output_shared_object_filename())

        # if we aren't a release the built object will still be named like foo.1.2.3, so we gotta
        # snag it by globbing since we dont know
        built_objects = list(Path(build_path).glob("libscrapli.*"))
        if len(built_objects) != 1:
            raise OSError("expecting exactly one built object...")

        return str(built_objects[0])

    @classmethod
    def _get_libscrapli_asset_filename(cls) -> str:
        zig_triple = os.environ.get(LIBSCRPALI_ZIG_TRIPLE_ENV, None)
        if zig_triple is not None:
            ext = "so" if "linux" in zig_triple else "dylib"
            return f"libscrapli-{zig_triple}.{ext}.{LIBSCRAPLI_VERSION}"

        _base = f"libscrapli-{cls._get_zig_style_arch()}"

        if sys.platform == "linux":
            lib_filename = f"{_base}-{cls._get_zig_style_platform()}.so.{LIBSCRAPLI_VERSION}"
        elif sys.platform == "darwin":
            lib_filename = f"{_base}-macos.dylib.{LIBSCRAPLI_VERSION}"
        else:
            raise NotImplementedError("unsupported platform")

        return lib_filename

    def _get_clone_command(self, tmp_build_dir: str) -> list[list[str]]:
        if self.libscrapli_is_release_tag:
            return [
                [
                    "git",
                    "clone",
                    "--branch",
                    self.libscrapli_tag,
                    "--depth",
                    "1",
                    "--single-branch",
                    LIBSCRAPLI_REPO,
                    tmp_build_dir,
                ]
            ]

        if self.libscrapli_tag == "":
            # unset means pull main
            return [
                [
                    "git",
                    "clone",
                    "--depth",
                    "1",
                    LIBSCRAPLI_REPO,
                    tmp_build_dir,
                ]
            ]

        if re.fullmatch(r"[0-9a-f]{7,40}", self.libscrapli_tag):
            # specific hash not main, not a release
            return [
                [
                    "git",
                    "clone",
                    LIBSCRAPLI_REPO,
                    tmp_build_dir,
                ],
                [
                    "git",
                    "-C",
                    tmp_build_dir,
                    "checkout",
                    self.libscrapli_tag,
                ],
            ]

        raise OSError("libscrapli tag is not a release tag, a hash, or empty (for build from main)")

    @staticmethod
    def _get_build_command(all_targets: bool = False) -> list[str]:
        cmd = [
            sys.executable,
            "-m",
            "ziglang",
            "build",
            "-freference-trace",
            "--summary",
            "all",
            "--",
            "--release",  # release mode
        ]

        if all_targets:
            cmd.append("--all-targets")

        return cmd

    def _download_asset(self, lib_dest: Path) -> None:
        _base = f"{LIBSCRAPLI_REPO}/releases/download/{self.libscrapli_tag}"
        _filename = self._get_libscrapli_asset_filename()
        asset_url = f"{_base}/{_filename}"

        with urlopen(asset_url) as response:
            if response.status != HTTPStatus.OK:
                raise OSError(f"failed downloading asset file at url '{asset_url}'")
            with open(lib_dest, mode="wb") as f:
                shutil.copyfileobj(response, f)

    def _build_from_source(self, lib_dest: Path) -> None:
        if LIBSCRPALI_ZIG_TRIPLE_ENV in os.environ:
            zig_triple = os.environ[LIBSCRPALI_ZIG_TRIPLE_ENV]
            wheel_platform = WHEEL_TARGETS[zig_triple]

            # we *technically* have a pure python package meaning we can lie to bdist_wheel and tell
            # it what our platform should be. so do that. we are "pure" python because we have no
            #  c-extensions, only the pre compiled zig shared object (i.e. .so or .dylib)
            self.plat_name_supplied = True
            self.plat_name = wheel_platform
        else:
            # not set in env, so not in ci, just do native/current platform like normal, but we
            # still need to know what that is so we grab the correct built shared object
            zig_triple = "native"

        libscrapli_build_path = self._get_libscrapli_build_path()

        if not Path(f"{libscrapli_build_path}/build.zig").exists():
            clone_commands = self._get_clone_command(tmp_build_dir=str(libscrapli_build_path))

            for clone_command in clone_commands:
                subprocess.check_call(clone_command)

            subprocess.check_call(
                # if no zig triple specified, then we are building for the current platform, so
                # don't build all the things
                self._get_build_command(all_targets=LIBSCRPALI_ZIG_TRIPLE_ENV in os.environ),
                cwd=libscrapli_build_path,
            )

        built_lib = self._get_libscrapli_built_shared_object_filename(
            build_path=f"{libscrapli_build_path}/zig-out/{zig_triple}/"
        )

        shutil.copy2(built_lib, lib_dest)

        self._clean_libscrapli_build_path(libscrapli_build_path=libscrapli_build_path)

    def _run_sdist_and_editable_wheel(self) -> None:
        src_lib_path = Path("scrapli/lib")
        out_lib_name = self._get_libscrapli_output_shared_object_filename()
        lib_dest = src_lib_path / out_lib_name

        if not LIBSCRAPLI_ALWAYS_BUILD and Path(src_lib_path / out_lib_name).exists():
            # requested target already exists, dont waste time building
            return

        if not LIBSCRAPLI_ALWAYS_BUILD and self.libscrapli_is_release_tag:
            self._download_asset(lib_dest=lib_dest)
        else:
            self._build_from_source(lib_dest=lib_dest)


class LibscrapliSdist(sdist, Libscrapli):  # type: ignore
    """Setuptools sdist dispatcher"""

    def run(self) -> None:
        """Run the sdist install"""
        self._run_sdist_and_editable_wheel()
        super().run()


class LibscrapliEditableWheel(editable_wheel, Libscrapli):  # type: ignore
    """Setuptools editable wheel dispatcher"""

    def run(self) -> None:
        """Run the editable wheel install"""
        self._run_sdist_and_editable_wheel()
        super().run()


class LibscrapliBdist(bdist_wheel, Libscrapli):  # type: ignore
    """Setuptools bdist dispatcher"""

    def run(self) -> None:
        """Run the bdist install"""
        if LIBSCRPALI_ZIG_TRIPLE_ENV in os.environ:
            zig_triple = os.environ[LIBSCRPALI_ZIG_TRIPLE_ENV]
            wheel_platform = WHEEL_TARGETS[zig_triple]

            # we *technically* have a pure python package meaning we can lie to bdist_wheel and tell
            # it what our platform should be. so do that. we are "pure" python because we have no
            #  c-extensions, only the pre compiled zig shared object (i.e. .so or .dylib)
            self.plat_name_supplied = True
            self.plat_name = wheel_platform
        else:
            # not set in env, so not in ci, just do native/current platform like normal, but we
            # still need to know what that is so we grab the correct built shared object
            zig_triple = "native"

        src_lib_path = Path("scrapli/lib")
        out_lib_name = self._get_libscrapli_output_shared_object_filename()
        lib_dest = src_lib_path / out_lib_name

        if not LIBSCRAPLI_ALWAYS_BUILD and Path(src_lib_path / out_lib_name).exists():
            # requested target already exists, dont waste time building
            super().run()

            return
        elif not LIBSCRAPLI_ALWAYS_BUILD and self.libscrapli_is_release_tag:
            self._download_asset(lib_dest=lib_dest)
        else:
            self._build_from_source(lib_dest=lib_dest)

        super().run()


setup(
    cmdclass={
        "sdist": LibscrapliSdist,
        "editable_wheel": LibscrapliEditableWheel,
        "bdist_wheel": LibscrapliBdist,
    },
)



"""cli_parse"""

import urllib.request
from importlib import import_module, resources
from io import BytesIO, TextIOWrapper
from logging import getLogger
from typing import Any, TextIO

from scrapli.exceptions import ParsingException

logger = getLogger(__name__)


def textfsm_get_template(platform: str, command: str) -> TextIO | None:
    """
    Find correct TextFSM template based on platform and command executed

    Args:
        platform: ntc-templates device type; i.e. cisco_ios, arista_eos, etc.
        command: string of command that was executed (to find appropriate template)

    Returns:
        None or TextIO of opened template

    Raises:
        N/A

    """
    try:
        import_module(name=".templates", package="ntc_templates")
        cli_table_obj = getattr(import_module(name=".clitable", package="textfsm"), "CliTable")
    except ModuleNotFoundError as exc:
        raise ParsingException("optional extra 'textfsm' not found") from exc

    template_dir = f"{resources.files('ntc_templates')}/templates"

    cli_table = cli_table_obj("index", template_dir)
    template_index = cli_table.index.GetRowMatch({"Platform": platform, "Command": command})

    if not template_index:
        logger.warning(
            f"No match in ntc_templates index for platform `{platform}` and command `{command}`"
        )
        return None

    template_name = cli_table.index.index[template_index]["Template"]

    return open(f"{template_dir}/{template_name}", encoding="utf-8")


def _textfsm_to_dict(
    structured_output: list[Any] | dict[str, Any], header: list[str]
) -> list[Any] | dict[str, Any]:
    """
    Create list of dicts from textfsm output and header

    Args:
        structured_output: parsed textfsm output
        header: list of strings representing column headers for textfsm output

    Returns:
        output: structured data

    Raises:
        N/A

    """
    logger.debug("converting textfsm output to dictionary representation")

    header_lower = [h.lower() for h in header]
    structured_output = [dict(zip(header_lower, row)) for row in structured_output]

    return structured_output


def textfsm_parse(
    template: str | TextIO, output: str, to_dict: bool = True
) -> list[Any] | dict[str, Any]:
    """
    Parse output with TextFSM and ntc-templates, try to return structured output

    Args:
        template: TextIO or string of URL or filesystem path to template to use to parse data
        output: unstructured output from device to parse
        to_dict: convert textfsm output from list of lists to list of dicts -- basically create dict
            from header and row data so it is easier to read/parse the output

    Returns:
        output: structured data

    Raises:
        ScrapliException: If raise_err is set and a textfsm parsing error occurs, raises from the
            originating textfsm.parser.TextFSMError exception.

    """
    import textfsm

    if isinstance(template, str):
        if template.startswith("http://") or template.startswith("https://"):
            with urllib.request.urlopen(template) as response:
                re_table = textfsm.TextFSM(
                    TextIOWrapper(
                        BytesIO(response.read()),
                        encoding=response.headers.get_content_charset(),
                    )
                )
        else:
            re_table = textfsm.TextFSM(open(template, mode="rb"))
    else:
        re_table = textfsm.TextFSM(template)

    try:
        structured_output: list[Any] | dict[str, Any] = re_table.ParseText(output)

        if to_dict:
            structured_output = _textfsm_to_dict(
                structured_output=structured_output, header=re_table.header
            )

        return structured_output
    except textfsm.parser.TextFSMError as exc:
        raise ParsingException("failed parsing output with 'textfsm'") from exc

    return []


def genie_parse(platform: str, command: str, output: str) -> list[Any] | dict[str, Any]:
    """
    Parse output with Cisco genie parsers, try to return structured output

    Args:
        platform: genie device type; i.e. iosxe, iosxr, etc.
        command: string of command that was executed (to find appropriate parser)
        output: unstructured output from device to parse

    Returns:
        output: structured data

    Raises:
        N/A

    """
    try:
        device = getattr(import_module(name=".conf.base", package="genie"), "Device")
        get_parser = getattr(
            import_module(name=".libs.parser.utils", package="genie"), "get_parser"
        )
    except ModuleNotFoundError as exc:
        raise ParsingException("optional extra 'genie' not found") from exc

    genie_device = device("scrapli_device", custom={"abstraction": {"order": ["os"]}}, os=platform)

    try:
        get_parser(command, genie_device)
        genie_parsed_result = genie_device.parse(command, output=output)
        if isinstance(genie_parsed_result, list | dict):
            return genie_parsed_result
    except Exception as exc:
        raise ParsingException("failed parsing output with 'genie'") from exc

    return []
